import json
import time
import websockets
import asyncio
import logging
from typing import Union, Optional, NoReturn
from .exceptions import CBotResponceError , CBotError
from .order import COrder
from .interface import IBotStore

class CBot:
    """Класс описывает объект простого бота, работающего с биржей Deribit.
    Запуск через метод run либо асинхронно через start.
    Бизнес логика самого бота описана в методе worker."""
    
    def __init__(self, url, auth:dict, instrument:dict, logger:Union[logging.Logger,str,None] = None, store:Optional[IBotStore] = None, order_label:str = 'dsb'):

        self.store = store
        self.url = url
        self.order_label = order_label
        self.instrument = {}
        self.order = {}
        self.__auth = None
        self.__credentials = auth
        self.logger = (logging.getLogger(logger) if isinstance(logger,str) else logger)

        if self.logger is None:
            self.logger = logging.getLogger(__name__)

        _d = instrument['default']

        for i in instrument:

            _i = i.upper()

            if _i == 'DEFAULT':
                continue

            self.instrument[_i] = _d if instrument[i] is None else { **_d, **instrument[i] }

        self.logger.debug(f'Bot init instrument = {self.instrument}')

    def create_message(self, method: str, params: dict = {}, id: Union[int, str, None] = None, as_dict:bool = False) -> Union[str,dict] :
        """Метод возвращает объект или строку дамп JSON с телом для запроса на API биржи"""
        obj = {
            "jsonrpc" : "2.0",
            "id" : ( str(time.time()).replace('.','_') if id is None else id ),
            "method" : method,
            "params" : params
        }

        self.logger.debug(f'Creat message = {obj}')

        return obj if as_dict else json.dumps(obj)


    def get_responce_result(self, raw_responce:str, raise_error:bool = True, result_prop:str = 'result') -> Optional[dict]:
        """Получает тело ответа от сервера, после чего возвращает объект находящийся в result_prop, либо 
        бросает исключение, если с сервера пришли данные об ошибке
        """

        obj = json.loads(raw_responce)

        self.logger.debug(f'Get response = {obj}')

        if result_prop in obj:
            return obj[result_prop]

        if 'error' in obj and raise_error:
            raise CBotResponceError(obj['error']['message'],obj['error']['code'])

        return None


    async def auth(self, ws) -> Optional[dict]:

        await ws.send(
            self.create_message(
                'public/auth',
                self.__credentials
            )
        )

        return self.get_responce_result(await ws.recv())


    async def create_order(self, direction:str, instrument:str, price:float, amount:float, order_prop:dict = {}, raise_error:bool = True) -> COrder:

        async with websockets.connect(self.url) as websocket:
            
            await self.auth(websocket)

            prop = {
                'instrument_name' : instrument,
                'amount' : amount,
                'type' : 'limit',
                'price' : price,
                'advanced': 'usd'
            }

            await websocket.send(
                self.create_message(
                    f'private/{direction}',
                    { **prop , **order_prop }
                )
            )

            order_res = self.get_responce_result(await websocket.recv(),raise_error = raise_error)

            if not (self.store is None) and not (order_res is None):
                self.store.insert(order_res['order'])

            return COrder(order = (None if order_res is None else order_res['order']))


    async def cancel_order(self, id:Union[int,str,None], active:bool = False, raise_error:bool = True) -> COrder:

        async with websockets.connect(self.url) as websocket:

            await self.auth(websocket)

            await websocket.send(
                self.create_message(
                    f'private/cancel',
                    { 'order_id': id }
                )
            )

            order_res = self.get_responce_result(await websocket.recv(),raise_error = raise_error)

            if not (self.store is None):
                self.store.update(
                    id = id, 
                    order = order_res, 
                    other_param = ({} if active else { 'active' : 0, 'active_comment' : 'Order cancelled'}), 
                    return_is_active = False
                )

            return COrder(order_res)


    async def get_order_state(self, id:Union[int,str], raise_error:bool = True) -> Optional[COrder]:

        async with websockets.connect(self.url) as websocket:

            await self.auth(websocket)

            await websocket.send(
                self.create_message(
                    f'private/get_order_state',
                    { 'order_id': id }
                )
            )

            return COrder(self.get_responce_result(await websocket.recv(), raise_error = raise_error))


    async def worker(self, instrument:str, gap:float, gap_ignore:float, amount:float, price_id:str = 'mark_price', order:COrder = None, prev_order:COrder = None) -> NoReturn:
        """Реализует логику работы бота"""

        if order is None:
            order = COrder()

        if prev_order is None:
            prev_order = COrder()

        group_id = order.label if not order.label is None else f'{self.order_label}.{instrument}.{time.time()}'

        is_change_order = False

        self.logger.info(f'{instrument} Start worker order = {order.id} , prev_order = {prev_order.id}, group_id = {group_id}')

        async with websockets.connect(self.url) as websocket:

            await self.auth(websocket)

            await websocket.send(
                self.create_message(
                    'private/subscribe',
                    {
                        "channels": [ 
                            f'ticker.{instrument}.raw', 
                            f'user.orders.{instrument}.raw' 
                        ]
                    }
                )
            )

            while websocket.open:

                # Учёт необходимости обновления данных об ордере
                if is_change_order and order.isset and not (self.store is None):
                    self.store.update(order.id, order.source)

                if order.state in ['rejected', 'cancelled', 'untriggered']:
                    self.logger.info(f'{instrument} reset order = {order.id} , state = {order.state}')
                    order.reset()

                # Получаем сообщение от сервера
                message = self.get_responce_result(await websocket.recv(), result_prop='params')
                data = None
                is_change_order = False

                if (
                        not (message is None) and 
                        ('channel' in message) and 
                        ('data' in message) 
                ):
                    data = message['data']
                else:
                    continue

                # Обработка сообщений приходящих от тикера
                if message['channel'].startswith('ticker') :

                    price = data[price_id]

                    if order.isset == False :

                        order = await self.create_order('buy', instrument, (price - gap / 2), amount, { 'trigger' : price_id, 'label' : group_id })
                        self.logger.info(f'{instrument} create {order.direction} order = {order.id}')

                    elif order.direction=='buy':
                        
                        if order.state == 'filled':
                            prev_order = order
                            order = await self.create_order('sell', instrument, (price + gap), order.amount, { 'trigger' : price_id, 'label' : group_id })
                            self.logger.info(f'{instrument} create {order.direction} order = {order.id}, price = {price}')

                        elif price > order.price + gap + gap_ignore :
                            await self.cancel_order(order.id)
                            self.logger.info(f'{instrument} cancel {order.direction} order = {order.id}, price = {price}, prev_order = {prev_order.id}')
                            order = prev_order

                    elif order.direction=='sell':

                        if order.state == 'filled':
                            prev_order = order
                            order = await self.create_order('buy', instrument, (price - gap / 2), order.amount, { 'trigger' : price_id, 'label' : group_id })
                            self.logger.info(f'{instrument} create {order.direction} order = {order.id}, price = {price}')

                        elif price < order.price - gap - gap_ignore :
                            await self.cancel_order(order.id)
                            self.logger.info(f'{instrument} cancel {order.direction} order = {order.id}, price = {price}, prev_order = {prev_order.id}')
                            order = prev_order

                # Обработка сообщений об изменении ордеров
                elif message['channel'].startswith('user.orders') :

                    if data['last_update_timestamp'] == data['creation_timestamp']:
                        continue

                    is_change_order = True

                    if data['order_id'] == order.id:
                        self.logger.info(f'{instrument} update {order.id}')
                        order.set_source(data)


    async def synch(self, mod:int = 1, get_actual:list[str] = []) -> dict:
        """Функция проводит синхронизацию ордеров из хранилища с биржей.
         
            - get_actual -  список инструментов, актуальных для текущего запуска, если он указан то:
                            по каждому из инструментов будут найдены актуальные ордера во внутреннем хранилище. 
                            После каждый из ордеров будет синхронизирован с биржей
            
            - mod -         режим синхронизации - указывает как надо проводить синхронизацию тех ордеров у которых
                            active = 1 (т.е. управляемые ботом). 

                            mod = 0 - ничего не делать
                            
                            mod = 1 - все ордера с active = 1 (кроме тех, которые актуальны для тех инструментов,
                            которые указаны в get_actual) имеющие статус Open на бирже - будут принудительно закрыты.
                            active примет значение 0

                            mod не 0 и не 1 - все ордера с active = 1 (кроме тех, которые актуальны для тех инструментов,
                            которые указаны в get_actual) получат active = 0, без доп.запросов на биржу
        """
        
        result = {}
        ignore_id = []

        self.logger.info(f'Start synch mod = {mod}, get_actual = {get_actual}')

        if len(get_actual):

            orders = self.store.get(
                param = { 'active' : 1 , 'instrument' : get_actual, 'state': ['open','filled'] }, 
                order_by= { 'real_create' : 'desc' }
            )
            
            for order in orders:

                _i = order['instrument']

                if (
                        (_i in result) and
                        (
                            (len(result[_i]) == 2) or 
                            (
                                len(result[_i])==1 and 
                                (result[_i][0].state == 'filled' or result[_i][0].label!=order['group_id'])
                            )
                        )
                ):
                    continue

                src_order = await self.get_order_state(order['id'], raise_error = False)

                if src_order.state in [None, 'rejected', 'cancelled', 'untriggered']:
                    continue

                self.store.update(order['id'],src_order.source)

                if not _i in result:
                    result[_i] = []

                result[_i].append(src_order)

                ignore_id.append(order['id'])


        if mod>0 : 

            other_param={ 'active' : 0 ,'active_comment' : 'Synch' }
            
            param = { 'active' : 1 }
            if len(ignore_id):
                param['id'] = { 'operation': 'not in', 'value':ignore_id }

            orders = self.store.get(param = param)

            for order in orders:

                if mod == 1:
                    if not order['state'] in ['open', 'filled']:
                        self.store.update(order['id'],other_param=other_param)
                    else:
                        src_order = await self.get_order_state(order['id'], raise_error = False)

                        if src_order.state == 'open':
                            await self.cancel_order(order['id'],raise_error = False)
                        else:
                            self.store.update(
                                order['id'], 
                                order = src_order.source, 
                                other_param=other_param
                            )
                else:
                    self.store.update(order['id'],other_param=other_param)

        return result


    async def start(self, synch_mod:int = 1, synch_actual:bool = True) -> NoReturn:
        """Стартует работу бота с указанием параметров для синхронизации.
        Синхронизация будет проводится только в случае если задано хранилище (store)"""

        self.logger.info(f'Bot start')

        order = {}
        task = []

        if not (self.store is None):
            order = await self.synch(
                mod = synch_mod,
                get_actual = ( list(self.instrument.keys()) if synch_actual else [] )
            )

        for i in self.instrument:
            
            _order = None
            _prev_order = None

            if i in order:
                _order = order[i][0]
                if len(order[i]) == 2:
                    _prev_order = order[i][1]

            task.append(
                asyncio.create_task(
                    self.worker(
                        instrument = i,
                        order = _order,
                        prev_order = _prev_order, 
                        **self.instrument[i]
                    )
                )
            )

        await asyncio.gather(*task)


    def run(self,synch_mod:int = 1, synch_actual:bool = True) -> NoReturn:
        """Объёртка для start для запуска без доп библиотек для управления асинхроном"""
        asyncio.get_event_loop().run_until_complete(self.start(synch_mod = synch_mod, synch_actual = synch_actual))


""" 'Интерфейс' (если быть точнее - жалкая ООП реплика на Python :) )
Сделана в основном, чтоб обозначить методы которыми должен обладать сторедж для хранения заявок.
Строгих требований к хранилищу нет. Данные хранение которых необходимо обеспечить и являются 
обязательными - это: 

 - id:[str,int]                 - order_id с биржи
 - state:[str]                  - order_state - с биржи: filled, open, reject и т.п.
 - instrument:str               - указывающий на инструмент ордера
 - group_id:[str]               - ID круппы ордера, вписывается в order.label на бирже
 - active:[int]                 - указывает на то, актуальный ли ордер или нет для бота 1 - да, 0 - нет
 - active_comment:[str]         - комментарий, в котором отмечается причина смены статуса active
 - real_create:[str:datetime]   - реальная дата создания объекта
 - operation:[str]              - направление операции bay или sell

Остальные параметры по желанию.
Порядок полей неважен.
Соотвественно объекты, возвращаемые методами должны содержать указанный набор данных.

Такой набор данных обеспечит синхронизацию с биржей после остановки.

Такой подход может обеспечить хранение данных где угодно.

Метод get должен позволять запрашивать запись по ID или по параметрам, указываемые следующим видом:

    - store.get(id = ...)   # Запрос по ID
    
    # Запрос ордеров у которых active = 1 и state = одному из значений ['open','fill']
    - store.get(
        param = {
            'active':1,
            'state': ['open','fill']
        }
    )

    # Запрос ордеров у которых active = 0 и instrument = 'BTC' и state любое НО НЕ ['open','fill']
    - store.get(
        param = {
            'active' : 1,
            'instrument' : 'BTC',
            'state' : {
                'operation' : ' not in ', 
                'value' : ['open','fill']
            }
        }
    )


"""

class IBotStore:

    def get(self, id = None, param = {}, order_by = {}):
        pass

    def insert(self, order, other_param, return_is_active = True, modify_active = True):
        pass

    def update(self, id, order, other_param, return_is_active = True, modify_active = True):
        pass


class CBotResponceError(Exception):
    def __init__(self, message, code):
        super().__init__(message)
        self.code = code

class CBotError(Exception):
    def __init__(self, message):
        super().__init__(message)


from typing import Optional

class COrder:

    def __init__(self, order:Optional[dict] = None):
        """Класс описывает объект который храниться в хранилище (описанный интерфейсом IBotStore). 
        Обеспечивает доступ к свойствам id, state, label и direction. В качестве аргумента принимает 
        объект (JSON) описывающий заказа на бирже - хранимый в свойстве source."""

        self.source = order

    def set_source(self, order:dict):
        self.source = order
        return self

    def reset(self):
        self.source = None
        return self

    def __getattr__(self,name):

        field_map = {
            'id' : 'order_id',
            'state' : 'order_state',
            'label' : 'label',
            'direction' : 'direction'
        }

        if name == 'source':
            return self.source
        elif name == 'isset':
            return (not (self.source is None))
        elif (name in field_map) and (self.source is None):
            return None
        elif (name in field_map):
            return self.source[field_map[name]]
        elif not (self.source is None) and (name in self.source):
            return self.source[name]
        else:
            raise AttributeError(f'{name} attribute not found in DerebitOrder')



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['deribitsimplebot']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'anyio>=2.1.0,<3.0.0',
 'json5>=0.9.5,<0.10.0',
 'mysql-connector-python>=8.0.23,<9.0.0',
 'websockets>=8.1,<9.0']

setup_kwargs = {
    'name': 'deribitsimplebot',
    'version': '0.2.2',
    'description': 'Class set for the implementation of a simple bot working with the Deribit crypto exchange',
    'long_description': '.. _header-n246:\n\nDeribitSimpleBot\n================\n\n| Реализация простого бота для криптобиржи Deribit.\\\\\n| Реализация для websockets JSON-RPC v.2\n  (`Документация <https://docs.deribit.com/>`__)\n\n.. _header-n249:\n\nВозможности\n-----------\n\n1. Возможность работы с несколькими инструментами одновременно\n\n2. Возобновление работы после остановки\n\n3. Возможность использвоать разные системы хранения ордеров\n\n.. _header-n257:\n\nАлгоримт работы\n---------------\n\n1. Робот выставляет ордер #1 на покупку по цене **buy price = current\n   price - gap / 2**.\n\n2. | **(a)** Если цена уменьшается до **buy price**, то ордер #1, скорее\n     всего, будет исполнен. В этом случае перейти к пункту 3.\\\\\n   | **(b)** Если цена увеличивается до такого значения, что становится\n     истинным условие **current price > buy price + gap + gap ignore**,\n     то робот должен отменить ордер #1. Далее, вернуться к пункту 1.\n\n3. Робот выставляет ордер #2 на продажу по цене **sell price = current\n   price + gap**.\n\n4. | **(a)** Если цена увеличивается до sell price, то ордер #2, скорее\n     всего, будет исполнен. В этом случае вернуться к пункту 1.\\\\\n   | **(b)** Если цена уменьшается до такого значения, что становится\n     истинным условие **current price < sell price - gap - gap ignore**,\n     то робот должен отменить ордер #2. После этого следует вернуться к\n     пункту 3.\n\n.. _header-n267:\n\nРабочий пример\n--------------\n\n`Пример\nприложение <https://github.com/n-eliseev/deribitsimplebot/tree/master/example-app>`__\n(`описание к нему <https://github.com/n-eliseev/deribitsimplebot/>`__)\n',
    'author': 'Eliseev Nikolay',
    'author_email': 'n.a.eliseev@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/n-eliseev/deribitsimplebot',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
Хранит реализацию двух классов для работы с MySQL:
 - CMySQLBotStore(IBotStore) - для хранения данных о заказах в БД
 - CLogMySQLHandler(logging.Handler) - для логирования в БД
"""

import json
import time
import logging
from .interface import IBotStore
from mysql.connector import MySQLConnection
from typing import Union, NoReturn


class CMySQLBotStore(IBotStore):
    """Реализация интерфейса IBotStore обеспечивающее хранение заявок в БД MySQL.
    Описание и требования см. IBotStore
    """

    def __init__(self,connection : MySQLConnection = None, **connection_option):
        """На входе надо передать, либо собранный коннектор, либо параметры для подключения
        через MySQLConnection (см. https://dev.mysql.com/doc/connector-python/en/connector-python-connectargs.html)
        """
        super().__init__()
        self.__connection = (MySQLConnection(**connection_option) if connection is None else connection) 
        self.__cursor = self.__connection.cursor(dictionary = True)

    def __del__(self):
        try:
            self.__cursor.close()
            self.__connection.close()
        except:
            pass

    def get(self, id:Union[int,str,None] = None, param:dict = {}, order_by:dict[str,str] = {}) -> Union[None, dict, list[dict]]:
        
        sql = 'select * from `order` where '
        _real_param = {}

        if not (id is None):
            
            sql+='`id`=%(id)s limit 1'
            _real_param = { 'id': id }

        elif len(param.keys()):

            _p = []
            
            for i in param:

                _cd = {
                    'operation' : '=',
                    'value' : param[i]
                }
 
                if isinstance(param[i],list):
                    _cd['operation'] = 'in'
                elif isinstance(param[i],dict):
                    _cd = param[i]

                if 'raw' in _cd:
                    _p.append(_cd['raw'])
                else:

                    if isinstance(_cd['value'],list):
                        
                        __p = []
                        
                        for j,v in enumerate(_cd['value']):
                            __p.append(f'%({i}_{j})s')
                            _real_param[f'{i}_{j}'] = v
                        
                        _pp = f'({",".join(__p)})'
                    
                    else:
                        _pp = f'%({i})s'
                        _real_param[i] = _cd['value']
                
                    _p.append(f'(`{i}` {_cd["operation"]} {_pp})')
    
            sql+=' and '.join(_p)

        if len(order_by):
            _ob=order_by.popitem()
            sql+=f' order by `{_ob[0]}` {_ob[1]}'

        self.__cursor.execute(sql,_real_param)
        order = self.__cursor.fetchone() if not (id is None) else self.__cursor.fetchall() 
        self.__connection.commit()

        return order


    def insert(self,order:dict, other_param:dict = {}, return_is_active:bool = True, modify_active:bool = True):
        return self.__write(
                is_insert = True, 
                order = order, 
                other_param=other_param,
                return_is_active = return_is_active,
                modify_active = modify_active,
                field_map = {
                    'id' : 'order_id',
                    'group_id' : 'label',
                    'instrument' : 'instrument_name',
                    'state' : 'order_state',
                    'type' : 'order_type',
                    'direction': 'direction',
                    'price' : 'price',
                    'amount' : 'amount',
                    'real_create' : 'creation_timestamp',
                    'raw_data' : 'raw_data'
                }
            )


    def update(self, id:Union[int,str,None], order:dict, other_param:dict = {}, return_is_active:bool = True, modify_active:bool = True) :
        return self.__write(
                is_insert = False, 
                id = id,
                order = order, 
                other_param=other_param,
                return_is_active = return_is_active,
                modify_active = modify_active,
                field_map = {
                    'state' : 'order_state',
                    'type' : 'order_type',
                    'price' : 'price',
                    'amount' : 'amount',
                    'update' : 'update',
                    'real_update' : 'last_update_timestamp',
                    'last_raw_data' : 'last_raw_data'
                }
            )


    def __write(self, is_insert:bool, field_map:dict, order:dict = None, id:Union[int,str,None] = None, other_param:dict = {}, return_is_active:bool = True, modify_active:bool = True):

        _order = { **other_param }

        if order is None:
            order = {}

        if 'active' in _order:
            modify_active = False

        for i in field_map :

            if (i == 'real_create' or i == 'real_update') and (field_map[i] in order):
                _order[i] = time.strftime('%Y-%m-%d %H:%M:%S',time.localtime(int(order[field_map[i]]/1000)))
            elif (i == 'raw_data') or (i == 'last_raw_data'):
                _order[i] = json.dumps(order)
            elif (i == 'update'):
                _order[i] = time.strftime('%Y-%m-%d %H:%M:%S',time.localtime())
            elif (field_map[i] in order):
                _order[i] = order[field_map[i]]
        
        if modify_active and (_order['state'] != 'open') and (_order['state'] != 'filled'):
            _order['active'] = 0
            _order['active_comment'] = 'Order is not open or filled'

        if is_insert:
            sql = f'insert into `order` (`{"`,`".join(_order.keys())}`) values (%({")s,%(".join(_order.keys())})s)'
        else :
            _p = [f'`{i}`=%({i})s' for i in _order.keys()]
            sql = f'update `order` set {", ".join(_p)} where `id` = %(id)s limit 1'
            _order['id'] = id

        self.__cursor.execute(sql,_order)
        self.__connection.commit()

        return self.get(param = {'id' : _order['id'],'active' : 1}) if return_is_active else self.get(_order['id'])



class CLogMySQLHandler(logging.Handler):
    """Реализация обработчика для записи логов в БД"""

    def __init__(self, connection : MySQLConnection = None, **connection_option):
        super().__init__()
        self.__connection = (MySQLConnection(**connection_option) if connection is None else connection)
        self.__cursor = self.__connection.cursor(dictionary = True)

    def emit(self, record):

        sql = 'insert into `log` (`sender_id`,`level`,`level_order`,`data`) values (%(sender_id)s,%(level)s,%(level_order)s,%(data)s)'

        param = {
            'sender_id' : record.name,
            'level' : record.levelname,
            'level_order' : record.levelno,
            'data' : record.msg
        }

        self.__cursor.execute(sql,param)
        self.__connection.commit()